/**
 * @file rdi_tunnel.cpp
 * @brief The RDI Tunneling application, allows the FasCam iOS application
 *        to communicate to the camera server via the USB tether.
 * 
 * @copyright Copyright (c) 2024
 */

//---------------------------------------------------------------------------
// Include Files
//---------------------------------------------------------------------------
#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <memory>
#include <string>

#include "rdi_tunnel.h"

//---------------------------------------------------------------------------
// Variable Declarations
//---------------------------------------------------------------------------

struct struct_rc rc;
struct struct_options options;
struct struct_settings settings = {0, 0, 0, 0, 0};

static struct option long_options[] =
{
    { "local-port",    required_argument, NULL, LOCAL_PORT_OPTION },
    { "tether-host",   required_argument, NULL, TETHER_HOST_OPTION },
    { "tether-port",   required_argument, NULL, TETHER_PORT_OPTION },
    { "log",           no_argument,       NULL, LOG_OPTION },
    { "local-retry",   no_argument,       NULL, LOCAL_RETRY_OPTION },
    { "help",          no_argument,       NULL, HELP_OPTION },
    { "version",       no_argument,       NULL, VERSION_OPTION },
    { 0, 0, 0, 0 }
};

/**
 * @brief The main entry point for the RDI Tunneling application.
 * 
 * @param argc  [in] Normal arguments as described in "help".
 * @param argv  [in] Normal arguments as described in "help".
 * @return Exit status: 0 on success, failure otherwise.
 */
int main(int argc, char *argv[])
{
    // Save the program name.
    name = argv[0];

    // Let folks know we've started.
    print_version();

    // Initialize our collection.
    memset(&rc, 0, sizeof(rc));
    rc.tether_socket = INVALID_SOCKET;
    rc.local_socket = INVALID_SOCKET;

    // Parse the command line settings.
    set_options(argc, argv);

    bool done = false;
    while (!done)
    {
        if (settings.log)
            printf("+++++++++++++++++++++++++++++++++++\n");

        // Connect to the tether interface.
        if (!build_tether())
        {
            // There was a problem connecting to the tether server.
            exit(1);
        }

        // Wait until we have data on the tether socket before we connect
        // to the local server.
        if (wait_for_tether())
        {
            // Build our connection to the local server.
            if (build_local())
            {
                // Start tunneling.
                use_tunnel();
            }
        }

        // All done. Close the socket to the tether server.
        if (rc.tether_socket != INVALID_SOCKET)
            close(rc.tether_socket);
        if (rc.local_socket != INVALID_SOCKET)
            close(rc.local_socket);
        rc.tether_socket = INVALID_SOCKET;
        rc.local_socket = INVALID_SOCKET;

        // Check to see if the device is still connected.
        // "/var/run/usbmuxd.pid"
        struct stat test_buf;
        if (stat("/var/run/usbmuxd.pid", &test_buf) != 0)
        {
            // We are done.
            if (settings.log)
                printf("Looks like the USBMUXD left the building...\n");
            done = true;
        }
        else
        {
            // Delay just a bit.
            sleep(1);
        }

        if (settings.log)
            printf("-----------------------------------\n");
    }

    // All is well.
    return 0;
}

void print_usage()
{
    fprintf(stderr, "Usage: %s [options]\n\n", name);
}

void print_helpinfo()
{
    fprintf(stderr, "Try `%s --help' for more options\n", name);
}

void print_help()
{
    fprintf(stderr, "\
Required:\n\
  --local-port=PORT     local port, typically 50051\n\
  --tether-port=PORT    tether port, typically 50052\n\
\n\
Optional:\n\
  --tether-host=HOST    tether host [default: 127.0.0.1]\n\
  --local-retry=RETRY   times to retry connecting to local host [default: %d]\n\
  --log\n\
  --version\n\
  --help\n\n\
\n", DEFAULT_RETRIES);
}

void print_version()
{
    fprintf(stderr, "%s: v%d.%d.%d\n\n", name, RT_MAJ, RT_MIN, RT_PATCH);
}

void print_missing(const char *message)
{
    print_usage();
    fprintf(stderr, "%s: %s\n", name, message);
    print_helpinfo();
}

void set_options(int argc, char *argv[])
{
    int opt;
    int index;

    // We need SOME arguments!!
    if (argc < 3)
    {
        print_usage();
        print_help();
        exit(1);
    }

    options.local_retry = DEFAULT_RETRIES;

    do
    {
        opt = getopt_long(argc, argv, "", long_options, &index);
        switch (opt)
        {
        case LOCAL_PORT_OPTION:
        {
            options.local_port = optarg;
            settings.local_port = 1;
            break;
        }

        case TETHER_PORT_OPTION:
        {
            options.tether_port = optarg;
            settings.tether_port = 1;
            break;
        }

        case TETHER_HOST_OPTION:
        {
            options.tether_host = optarg;
            settings.tether_host = 1;
            break;
        }

        case LOG_OPTION:
        {
            settings.log = 1;
            break;
        }

        case LOCAL_RETRY_OPTION:
        {
            options.local_retry = atoi(optarg);
            settings.local_retry = 1;
            break;
        }

        case HELP_OPTION:
        {
            print_usage();
            print_help();
            exit(0);
        }

        case VERSION_OPTION:
        {
            print_version();
            exit(0);
        }

        case '?':
        {
            print_usage();
            print_helpinfo();
            exit(0);
        }
        }
    } while (opt != -1);

    if (!settings.local_port)
    {
        print_missing("missing '--local-port=' option.");
        exit(1);
    }

    if (!settings.tether_port)
    {
        print_missing("missing '--tether-port=' option.");
        exit(1);
    }
}

/**
 * @brief Create the socket to the "tether_addr:tether_port" and connect to it.
 * 
 * @return True on success, false if the call failed.
 */
bool build_tether()
{
    const bool OK(true);

    if (settings.log)
        printf("%s\n", __FUNCTION__);

    // Clear out our address structure.
    memset(&rc.tether_addr, 0, sizeof(rc.tether_addr));

    rc.tether_addr.sin_family = AF_INET;
    rc.tether_addr.sin_port = htons(atoi(options.tether_port));

    // If the user didn't supply a host name, then we will be using the local host.
    // Otherwise, try to get information about the host the user asked for.
    if (!settings.tether_host)
    {
        rc.tether_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
    }
    else
    {
        // Get the information from the DNS.
        rc.tether_host = gethostbyname(options.tether_host);
        if (rc.tether_host == NULL)
        {
            perror("build_tether: gethostbyname()");
            return !OK;
        }

        memcpy(&rc.tether_addr.sin_addr.s_addr, rc.tether_host->h_addr, rc.tether_host->h_length);
    }

    // Create the socket to talk with.
    rc.tether_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (rc.tether_socket < 0)
    {
        perror("build_tether: socket()");
        return !OK;
    }
    if (settings.log)
        printf("%s: tether socket=%d:%d\n", __FUNCTION__, rc.tether_socket, atoi(options.tether_port));

    // See if we can connect to the tether server.
    if (connect(rc.tether_socket, (struct sockaddr *) &rc.tether_addr, sizeof(rc.tether_addr)) < 0)
    {
        perror("build_tether: connect()");
        return !OK;
    }

    if (settings.log)
        printf("%s: done and connected\n", __FUNCTION__);
    return OK;
}

/**
 * @brief Wait for network activity on the tether socket.
 * 
 * @return True if there is activity, false if an error forced us to exit.
 */
bool wait_for_tether()
{
    bool activity(false);       // Assume there will be an error.

    if (settings.log)
        printf("%s\n", __FUNCTION__);

    for (;;)
    {
        fd_set io;
        FD_ZERO(&io);
        FD_SET(rc.tether_socket, &io);

        // Wait for activity on the tether socket.
        if (select(fd(), &io, NULL, NULL, NULL) < 0)
        {
            perror("wait_for_tether: select()");
            break;
        }

        // Is our flag set?
        if (FD_ISSET(rc.tether_socket, &io))
        {
            if (settings.log)
                printf("%s: Got traffic on the tether.\n", __FUNCTION__);

            // We have activity!
            activity = true;
            break;
        }
    }

    if (settings.log)
        printf("%s: activity=%d\n", __FUNCTION__, activity);
    return activity;
}

/**
 * @brief Create the socket to the "local_addr:local_port" and connect to it.
 * 
 * @return True on success, false if the call failed.
 */
bool build_local()
{
    const bool OK(true);

    if (settings.log)
        printf("%s\n", __FUNCTION__);

    // Clear out our address structure.
    memset(&rc.local_addr, 0, sizeof(rc.local_addr));

    rc.local_addr.sin_family = AF_INET;
    rc.local_addr.sin_port = htons(atoi(options.local_port));
    rc.local_addr.sin_addr.s_addr = inet_addr("127.0.0.1");

    // Create the socket to talk with.
    rc.local_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (rc.local_socket < 0)
    {
        perror("build_local: socket()");
        return !OK;
    }
    if (settings.log)
        printf("%s: local socket=%d:%d\n", __FUNCTION__, rc.local_socket, atoi(options.local_port));

    // See if we can connect to the tether server.
    int i;
    for (i = 0; i < options.local_retry; ++i)
    {
        if (settings.log)
            printf("%s: Connect try %d\n", __FUNCTION__, i);

        if (connect(rc.local_socket, (struct sockaddr *) &rc.local_addr, sizeof(rc.local_addr)) < 0)
        {
            perror("build_local: connect()");

            // Wait for a little.
            sleep(1);
        }
        else
        {
            // We are connected.
            if (settings.log)
                printf("%s: connected\n", __FUNCTION__);
            break;
        }
    }
    if (i >= options.local_retry)
    {
        return !OK;
    }

    return OK;
}

/**
 * @brief Use the tunnel we have created between the tether and local sockets.
 * 
 * This is simply a pass-through, moving data from one side to the other.
 */
void use_tunnel()
{
    if (settings.log)
        printf("%s\n", __FUNCTION__);

    // Allocate our buffer.
    std::unique_ptr<uint8_t[]> buffer = std::make_unique<uint8_t[]>(DEFAULT_BUFFER_SIZE);
    if (buffer == NULL)
    {
        fprintf(stderr, "use_tunnel: Unable to allocate buffer!\n");
        return;
    }

    // We will loop until one side or the other disconnects.
    for (;;)
    {
        fd_set io;
        FD_ZERO(&io);
        FD_SET(rc.local_socket, &io);
        FD_SET(rc.tether_socket, &io);

        // Wait for activity on one side or the other.
        if (select(fd(), &io, NULL, NULL, NULL) < 0)
        {
            perror("use_tunnel: select()");
            break;
        }

        // Check for activity on the local socket.
        if (FD_ISSET(rc.local_socket, &io))
        {
            // Read the data from the local socket.
            ssize_t count = recv(rc.local_socket, buffer.get(), DEFAULT_BUFFER_SIZE, 0);
            if (settings.log)
                printf("%s: local_socket, count=%ld\n", __FUNCTION__, count);

            // Count will be <0 if there is an error.
            if (count < 0)
            {
                perror("use_tunnel: recv(rc.local_socket)");
                break;
            }

            // Count will be zero if the server has shutdown.
            if (count == 0)
            {
                if (settings.log)
                    printf("%s: local connection closed.\n", __FUNCTION__);
                break;
            }

            // Forward on the data to the tethered socket.
            if (settings.log)
                printf("%s: Forwarding %ld bytes to tether socket (%d)\n", __FUNCTION__, count, rc.tether_socket);
            send(rc.tether_socket, buffer.get(), count, 0);
        }

        // Check for activity on the tether socket.
        if (FD_ISSET(rc.tether_socket, &io))
        {
            ssize_t count = recv(rc.tether_socket, buffer.get(), DEFAULT_BUFFER_SIZE, 0);
            if (settings.log)
                printf("%s: tether_socket, count=%ld\n", __FUNCTION__, count);

            // Count will be <0 if there is an error.
            if (count < 0)
            {
                perror("use_tunnel: recv(rc.tether_socket)");
                break;
            }

            // Count will be zero if the server has shutdown.
            if (count == 0)
            {
                if (settings.log)
                    printf("%s: tether connection closed.\n", __FUNCTION__);
                break;
            }

            // Forward on the data to the local socket.
            if (settings.log)
                printf("%s: Forwarding %ld bytes to local socket (%d)\n", __FUNCTION__, count, rc.local_socket);
            send(rc.local_socket, buffer.get(), count, 0);
        }
    }

    if (settings.log)
        printf("%s: done\n", __FUNCTION__);
}

/**
 * @brief Utility function to find the largest file descriptor between
 *        the tether and local sockets. We increment the descriptor by
 *        one to have it ready for the FD_ISSET, et. al.
 * 
 * @return The value of the largest file descriptor.
 */
int fd()
{
    int fd = rc.local_socket;
    if (fd < rc.tether_socket)
    {
        fd = rc.tether_socket;
    }
    return fd + 1;
}
